from typing import Any

import pytest
from pydantic import RootModel, TypeAdapter

from formkit_ninja import models
from formkit_ninja.formkit_schema import (DiscriminatedNodeType, FormKitNode,
                                          FormKitSchemaFormKit, GroupNode,
                                          RepeaterNode, TextAreaNode, TextNode)
from tests.test_python_models import schema_are_same


class Node(RootModel):
    root: FormKitSchemaFormKit


type_adapter = TypeAdapter(FormKitSchemaFormKit)


def test_text():
    n = Node({"$formkit": "text"})
    n.model_dump(by_alias=True, exclude_none=True)


def test_group():
    n = Node({"$formkit": "group", "children": [{"$formkit": "text"}]})
    n.root.model_dump(by_alias=True, exclude_none=True)


def test_nested_group():
    n = Node.model_validate(
        {
            "$formkit": "group",
            "randomprop": "random",
            "children": [
                {
                    "$formkit": "group",
                    "children": [{"$formkit": "group", "randomprop": "random"}],
                }
            ],
        }
    )
    n.root.model_dump(by_alias=True, exclude_none=True)


def test_nested_group_adapter():
    n = GroupNode.model_validate({"$formkit": "group", "randomprop": "random"})
    print(n)


def test_node_type():
    assert isinstance(
        DiscriminatedNodeType.model_validate({"$formkit": "text"}).root, TextNode
    )
    assert isinstance(
        DiscriminatedNodeType.model_validate({"$formkit": "textarea"}).root,
        TextAreaNode,
    )


def test_group_disc():
    schema_in = {
        "$formkit": "group",
        "children": [{"$formkit": "text", "class": "red"}],
    }
    group = DiscriminatedNodeType.model_validate(schema_in)
    assert isinstance(group.root, GroupNode)
    assert isinstance(group.root.children[0], TextNode)
    assert group.root.children[0].additional_props == {"class": "red"}

    # When we do `group.model_dump` we get an error!
    schema_out = group.root.model_dump(by_alias=True, exclude_none=True)
    assert schema_in == schema_out


def test_group_disc_again():
    g = GroupNode()
    d = DiscriminatedNodeType(root=g)
    d.model_dump(by_alias=True, exclude_none=True)

    t = TextNode()
    g = GroupNode(children=[t])
    d = DiscriminatedNodeType(root=g)
    d.model_dump(by_alias=True, exclude_none=True)

    t = TextNode(additional_props={"class": "red"})
    g = GroupNode(children=[t])
    d = DiscriminatedNodeType(root=g)
    assert isinstance(d.root, GroupNode)
    assert isinstance(d.root.children[0], TextNode)
    assert d.root.children[0].additional_props == {"class": "red"}
    d.model_dump(by_alias=True, exclude_none=True)


def test_element():
    schema_in = {
        "$el": "button",
        "attrs": {
            "class": "disabled:hidden cursor-pointer flex items-center justify-center flex-row-reverse h-[50px] rounded-2.5xl px-4 w-full bg-white border-2 border-solid border-emerald-600 text-zinc-900 hover:bg-emerald-600 hover:text-white font-bold text-base",
            "data-index": "$index",
            "data-repeaterid": "$id",
            "onClick": "$attrs.removeAction",
        },
        "children": ["$attrs.removeLabel"],
        "if": "$value.length > 1",
    }

    group = DiscriminatedNodeType.model_validate(schema_in)
    assert group.model_dump(by_alias=True, exclude_none=True) == schema_in


def test_conditional():
    """
    Part of CFM 4
    """
    datepicker = {
        "$formkit": "datepicker",
        "_currentDate": "$getCurrentDate",
        "calendarIcon": "calendar",
        "format": "DD/MM/YYYY",
        "id": "date",
        "key": "date",
        "label": '$gettext("Date")',
        "name": "date",
        "nextIcon": "angleRight",
        "prevIcon": "angleLeft",
        "sectionsSchema": {
            "day": {
                "children": [
                    "$day.getDate()",
                    {
                        "children": [
                            {
                                "children": [
                                    {
                                        "$el": "div",
                                        "attrs": {"class": "formkit-day-highlight"},
                                        "if": "$attrs._currentDate().year === $day.getFullYear()",
                                    }
                                ],
                                "if": "$attrs._currentDate().month === $day.getMonth()",
                            }
                        ],
                        "if": "$attrs._currentDate().day === $day.getDate()",
                    },
                ]
            }
        },
    }

    group = DiscriminatedNodeType.model_validate(datepicker)
    assert group.model_dump(by_alias=True, exclude_none=True) == datepicker


simple_group = {
    "$formkit": "group",
    "children": [
        {
            "$formkit": "text",
            "key": "name",
            "label": "Name",
            "placeholder": "Enter your name",
        },
        {
            "$formkit": "text",
            "key": "email",
            "label": "Email",
            "placeholder": "Enter your email",
        },
    ],
}

uuid_schema = {"$formkit": "uuid", "name": "uuid", "readonly": True}


tel = {
    "$formkit": "tel",
    "label": '$gettext("Phone number")',
    "maxLength": 8,
    "name": "phone_number",
    "validation": "number|length:8,8",
}

sf_repeater = {
    "$formkit": "repeater",
    "addLabel": '$gettext("Add another member")',
    "children": [
        {
            "$el": "div",
            "attrs": {
                "class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"
            },
            "children": [
                {"$el": "span", "children": ["$: ($index + 1)"]},
                " Suku facilitator",
            ],
        },
        {"$formkit": "uuid", "name": "uuid", "readonly": True},
        #   {
        #    "$formkit": "select",
        #    "label": "$gettext(\"Round\")",
        #    "name": "round",
        #    "options":  "$ida(electionround)",
        #    "placeholder": "$gettext(\"Please select\")"
        #   },
        #   {
        #    "$formkit": "select",
        #    "label": "$gettext(\"Position\")",
        #    "name": "position",
        #    "options": "$ida(teamposition, \"group_id=2\")",
        #    "placeholder": "$gettext(\"Please select\")"
        #   },
        #   {
        #    "$formkit": "text",
        #    "label": "$gettext(\"Name\")",
        #    "name": "name",
        #    "placeholder": "$gettext(\"Please enter\")"
        #   },
        #   {
        #    "$formkit": "select",
        #    "label": "$gettext(\"Gender\")",
        #    "name": "gender",
        #    "options": "$ida(gender)",
        #    "placeholder": "$gettext(\"Please select\")"
        #   },
        #   {
        #    "$formkit": "select",
        #    "label": "$gettext(\"Person with a disability\")",
        #    "name": "person_with_disability",
        #    "options": "$ida(yesno)",
        #    "placeholder": "$gettext(\"Please select\")"
        #   },
        #   {
        #    "$formkit": "tel",
        #    "label": "$gettext(\"Phone number\")",
        #    "maxLength": 8,
        #    "name": "phone_number",
        #    "validation": "number|length:8,8"
        #   },
        #   {
        #    "$formkit": "date",
        #    "label": "$gettext(\"Date of exit from the committee\")",
        #    "name": "date_exit_committee"
        #   },
        #   {
        #    "$formkit": "text",
        #    "label": "$gettext(\"Reason for exit\")",
        #    "name": "exit_details"
        #   },
        #   {
        #    "$formkit": "select",
        #    "label": "$gettext(\"Is active?\")",
        #    "name": "active_status",
        #    "options": "$ida(yesno)",
        #    "value": "1"
        #   }
    ],
    "downControl": False,
    "id": "repeaterSukus",
    "itemClass": "repeater-children-index",
    "itemsClass": "repeater",
    "max": 2,
    "min": 1,
    "name": "repeaterSukus",
    "removeAction": "$repeaterRemoveAction",
    "removeLabel": '$gettext("Remove")',
    "sectionsSchema": {
        "remove": {
            "children": [
                {
                    "$el": "button",
                    "attrs": {
                        "class": "disabled:hidden cursor-pointer flex items-center justify-center flex-row-reverse h-[50px] rounded-2.5xl px-4 w-full bg-white border-2 border-solid border-emerald-600 text-zinc-900 hover:bg-emerald-600 hover:text-white font-bold text-base",
                        "data-index": "$index",
                        "data-repeaterid": "$id",
                        "onClick": "$attrs.removeAction",
                    },
                    "children": ["$attrs.removeLabel"],
                    "if": "$value.length > 1",
                }
            ]
        }
    },
    "upControl": False,
    "validationRules": "validateSukuFacilitators",
}


sf_41_datpicker = {
    "$formkit": "datepicker",
    "_currentDate": "$getCurrentDate",
    "calendarIcon": "calendar",
    "format": "DD/MM/YYYY",
    "id": "date",
    "key": "date",
    "label": '$gettext("Date")',
    "name": "date",
    "nextIcon": "angleRight",
    "prevIcon": "angleLeft",
    "sectionsSchema": {
        "day": {
            "children": [
                "$day.getDate()",
                {
                    "children": [
                        {
                            "children": [
                                {
                                    "$el": "div",
                                    "attrs": {"class": "formkit-day-highlight"},
                                    "if": "$attrs._currentDate().year === $day.getFullYear()",
                                }
                            ],
                            "if": "$attrs._currentDate().month === $day.getMonth()",
                        }
                    ],
                    "if": "$attrs._currentDate().day === $day.getDate()",
                },
            ]
        }
    },
}

sf_41_district = {
    "$formkit": "select",
    "id": "district",
    "key": "district",
    "label": "$gettext(Municipality)",
    "name": "district",
    "options": "$getLocations()",
}

sf_41_locations = {
    "$formkit": "select",
    "id": "administrative_post",
    "if": "$get(district).value",
    "key": "administrative_post",
    "label": '$gettext("Administrative Post")',
    "name": "administrative_post",
    "options": "$getLocations($get(district).value)",
}

af_41_sucu = {
    "$formkit": "select",
    "id": "suco",
    "if": "$get(administrative_post).value",
    "key": "suco",
    "label": "$gettext(Suco)",
    "name": "suco",
    "options": "$getLocations($get(district).value, $get(administrative_post).value)",
}
sf_41_activityname = {
    "$formkit": "select",
    "id": "activity",
    "key": "activity",
    "label": '$gettext("Activity name")',
    "name": "activity",
    "options": "$ida(cycle)",
    "placeholder": '$gettext("Please select")',
}
sf_41_meeting = {
    "$formkit": "dropdown",
    "key": "meeting_objective",
    "label": '$gettext("Meeting objective")',
    "name": "meeting_objective",
    "options": ["First accountability meeting", "Final accountability meeting2"],
    "selectIcon": "angleDown",
}

repeater = {
    "$formkit": "repeater",
    "addLabel": '$gettext("Add project")',
}

repeater_2 = {
    "$formkit": "repeater",
    "children": [
        {
            "$formkit": "repeater",
        },
        {
            "$formkit": "repeater",
            "children": [{"$formkit": "repeater", "children": [{"$formkit": "text"}]}],
        },
    ],
}

div_el = {
    "$el": "div",
    "attrs": {"class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"},
    "children": ["Output ", {"$el": "span", "children": ["$: ($index + 1)"]}],
}


@pytest.mark.parametrize(
    "schema",
    [
        simple_group,
        uuid_schema,
        tel,
        sf_repeater,
        sf_41_activityname,
        sf_41_datpicker,
        sf_41_district,
        sf_41_locations,
        af_41_sucu,
        sf_41_meeting,
        repeater,
        repeater_2,
        div_el,
    ],
)
def test_param_schemas(schema: dict[str, Any] | dict[str, str]):
    model_validated = DiscriminatedNodeType.model_validate(schema)

    # We get strange things happening
    schema_out = model_validated.root.model_dump(by_alias=True, exclude_none=True)
    assert schema_out == schema


def test_el():
    div_el = {
        "$el": "div",
        "attrs": {"class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"},
        "children": ["Output ", {"$el": "span", "children": ["$: ($index + 1)"]}],
    }

    model_validated = DiscriminatedNodeType.model_validate(div_el)
    # We get strange things happening
    schema_out = model_validated.model_dump(by_alias=True, exclude_none=True)
    schema_are_same(schema_out, div_el)


@pytest.mark.django_db
@pytest.mark.parametrize(
    "schema",
    [
        simple_group,
        uuid_schema,
        tel,
        sf_repeater,
        sf_41_activityname,
        sf_41_datpicker,
        sf_41_district,
        sf_41_locations,
        af_41_sucu,
        sf_41_meeting,
        repeater,
        repeater_2,
        div_el,
    ],
)
def test_roundtrip_on_database(schema: dict):
    m = DiscriminatedNodeType.model_validate(schema)
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(m.root))[0]
    schema_from_db: FormKitNode = node_in_the_db.to_pydantic(
        recursive=True, options=True
    )

    schema_out = schema_from_db.model_dump(by_alias=True, exclude_none=True)

    DiscriminatedNodeType.model_validate(node_in_the_db.get_node_values())

    schema_are_same(schema_out, schema)


def test_formkit_value_on_children():
    """
    This is an error case where "$formkit" is not correctly set on "child" nodes
    """
    t = TextNode()
    g = GroupNode(children=[t])
    f = FormKitNode(root=g)
    value = f.model_dump(by_alias=True, exclude_none=True, exclude_defaults=False)

    assert t.formkit == "text"
    assert t.node_type == "formkit"
    assert f.root.children[0].model_dump(by_alias=True, exclude_none=True) == {
        "$formkit": "text"
    }
    assert f.root.children[0].formkit == "text"
    assert "$formkit" in value["children"][0]
    assert value["children"][0]["$formkit"] == "text"


def test_repeater():
    r = RepeaterNode()
    t = TextNode()
    r.children = [t]
    assert r.model_dump(by_alias=True, exclude_none=True)


def test_repeater_prop():
    r = RepeaterNode(upControl=False)
    assert r.model_dump(by_alias=True, exclude_none=True)



# Generated by Django 4.2.6 on 2023-10-25 18:05

import uuid

import django.db.models.deletion
import pgtrigger.compiler
import pgtrigger.migrations
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("pghistory", "0006_delete_aggregateevent"),
        ("formkit_ninja", "0022_formkitschemanode_soft_delete"),
    ]

    operations = [
        migrations.CreateModel(
            name="FormKitSchemaNodeEvent",
            fields=[
                ("pgh_id", models.AutoField(primary_key=True, serialize=False)),
                ("pgh_created_at", models.DateTimeField(auto_now_add=True)),
                ("pgh_label", models.TextField(help_text="The event label.")),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "node_type",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("$cmp", "Component"),
                            ("text", "Text"),
                            ("condition", "Condition"),
                            ("$formkit", "FormKit"),
                            ("$el", "Element"),
                            ("raw", "Raw JSON"),
                        ],
                        max_length=256,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        blank=True,
                        help_text="Decribe the type of data / reason for this component",
                        max_length=4000,
                        null=True,
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        blank=True,
                        help_text="Used as a human-readable label",
                        max_length=1024,
                        null=True,
                    ),
                ),
                ("is_active", models.BooleanField(default=True)),
                (
                    "node",
                    models.JSONField(
                        blank=True,
                        help_text="A JSON representation of select parts of the FormKit schema",
                        null=True,
                    ),
                ),
                (
                    "additional_props",
                    models.JSONField(
                        blank=True,
                        help_text="User space for additional, less used props",
                        null=True,
                    ),
                ),
                (
                    "text_content",
                    models.TextField(
                        blank=True,
                        help_text="Content for a text element, for children of an $el type component",
                        null=True,
                    ),
                ),
                ("track_change", models.BigIntegerField(blank=True, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="insert_insert",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='INSERT INTO "formkit_ninja_formkitschemanodeevent" ("additional_props", "created", "created_by_id", "description", "id", "is_active", "label", "node", "node_type", "option_group_id", "pgh_context_id", "pgh_created_at", "pgh_label", "pgh_obj_id", "text_content", "track_change", "updated", "updated_by_id") VALUES (NEW."additional_props", NEW."created", NEW."created_by_id", NEW."description", NEW."id", NEW."is_active", NEW."label", NEW."node", NEW."node_type", NEW."option_group_id", _pgh_attach_context(), NOW(), \'insert\', NEW."id", NEW."text_content", NEW."track_change", NEW."updated", NEW."updated_by_id"); RETURN NULL;',
                    hash="534b4b21d7cf93a9105cdc2c791147f3365be30d",
                    operation="INSERT",
                    pgid="pgtrigger_insert_insert_72c4c",
                    table="formkit_ninja_formkitschemanode",
                    when="AFTER",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="update_update",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (OLD.* IS DISTINCT FROM NEW.*)",
                    func='INSERT INTO "formkit_ninja_formkitschemanodeevent" ("additional_props", "created", "created_by_id", "description", "id", "is_active", "label", "node", "node_type", "option_group_id", "pgh_context_id", "pgh_created_at", "pgh_label", "pgh_obj_id", "text_content", "track_change", "updated", "updated_by_id") VALUES (NEW."additional_props", NEW."created", NEW."created_by_id", NEW."description", NEW."id", NEW."is_active", NEW."label", NEW."node", NEW."node_type", NEW."option_group_id", _pgh_attach_context(), NOW(), \'update\', NEW."id", NEW."text_content", NEW."track_change", NEW."updated", NEW."updated_by_id"); RETURN NULL;',
                    hash="0154bfb8e2197684b693fce5fec3631e30fa5537",
                    operation="UPDATE",
                    pgid="pgtrigger_update_update_d7c99",
                    table="formkit_ninja_formkitschemanode",
                    when="AFTER",
                ),
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanodeevent",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                related_query_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanodeevent",
            name="option_group",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                related_query_name="+",
                to="formkit_ninja.optiongroup",
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanodeevent",
            name="pgh_context",
            field=models.ForeignKey(
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="pghistory.context",
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanodeevent",
            name="pgh_obj",
            field=models.ForeignKey(
                db_constraint=False,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="events",
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanodeevent",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                related_query_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django.core.management.base import BaseCommand

from formkit_ninja.models import FormKitSchemaNode, check_valid_django_id


class Command(BaseCommand):
    help = "Check node 'name' fields to be valid python identifiers"

    def handle(self, *args, **options):
        for node in FormKitSchemaNode.objects.all():
            if node.node and "name" in node.node:
                name = node.node["name"]
                try:
                    check_valid_django_id(name)
                    # self.stdout.write(self.style.SUCCESS(name))
                except TypeError:
                    self.stdout.write(self.style.WARNING(f"{node.pk}: {name}"))



import importlib
import re
from functools import cached_property
from http import HTTPStatus
from importlib.util import find_spec
from types import ModuleType
from typing import Sequence
from uuid import UUID, uuid4

from django.db import transaction
from django.db.models import F
from django.db.models.aggregates import Max
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.cache import add_never_cache_headers
from ninja import Field, ModelSchema, Router, Schema
from pydantic import BaseModel, ConfigDict

from formkit_ninja import formkit_schema, models

if find_spec("sentry_sdk"):
    sentry_sdk: ModuleType | None = importlib.import_module("sentry_sdk")
else:
    sentry_sdk = None


def sentry_message(message: str):
    if sentry_sdk and hasattr(sentry_sdk, "capture_message"):
        sentry_sdk.capture_message(f"{message}")


router = Router(tags=["FormKit"])


class FormKitSchemaIn(ModelSchema):
    class Meta:
        model = models.FormKitSchema
        fields = "__all__"


class SchemaLabel(ModelSchema):
    class Meta:
        model = models.SchemaLabel
        fields = ["lang", "label"]


class SchemaDescription(ModelSchema):
    class Meta:
        model = models.SchemaDescription
        fields = ["lang", "label"]


class FormKitSchemaListOut(ModelSchema):
    schemalabel_set: list[SchemaLabel]
    schemadescription_set: list[SchemaDescription]

    class Meta:
        model = models.FormKitSchema
        fields = ["id", "label"]


class FormComponentsOut(ModelSchema):
    node_id: UUID
    schema_id: UUID

    class Meta:
        model = models.FormComponents
        fields = ["label"]


class NodeChildrenOut(ModelSchema):
    children: list[UUID] = []
    latest_change: int | None = None

    class Meta:
        model = models.NodeChildren
        fields = ["parent"]


class NodeReturnType(BaseModel):
    key: UUID
    last_updated: int
    node: formkit_schema.FormKitNode
    protected: bool


class NodeInactiveType(BaseModel):
    key: UUID
    last_updated: int
    is_active: bool = False
    protected: bool


class NodeStringType(NodeReturnType):
    """
    str | formkit_schema.FormKitNode causes openapi generator to fail
    """

    node: str


NodeQSResponse = Sequence[NodeStringType | NodeReturnType | NodeInactiveType]


def node_queryset_response(qs: models.NodeQS) -> NodeQSResponse:
    responses = []
    n: NodeStringType | NodeInactiveType | NodeReturnType
    for key, last_updated, node, protected in qs.to_response(ignore_errors=False):
        if isinstance(node, str):
            n = NodeStringType(
                key=key, last_updated=last_updated, protected=protected, node=node
            )
        elif node is None:
            n = NodeInactiveType(
                key=key, last_updated=last_updated, protected=protected, is_active=False
            )
        else:
            n = NodeReturnType(
                key=key, last_updated=last_updated, protected=protected, node=node
            )
        responses.append(n)
    return responses


class Option(ModelSchema):
    group_name: str  # This is annotation of the model `content_type_model`
    value: str
    # Note: For other projects you may want to extend this with additional languages
    label_tet: str | None
    label_en: str | None
    label_pt: str | None
    # This is an optional field used to indicate the last update
    # It's linked to a Django pg trigger instance in Partisipa
    change_id: int | None = None

    class Meta:
        model = models.Option
        fields = ["value"]


@router.get("list-schemas", response=list[FormKitSchemaListOut])
def get_list_schemas(request):
    return models.FormKitSchema.objects.all()


@router.get("list-nodes", response=NodeQSResponse, by_alias=True, exclude_none=True)
def get_formkit_nodes(
    request: HttpRequest, response: HttpResponse, latest_change: int | None = -1
):
    """
    Get all of the FormKit nodes in the database
    """
    objects: models.NodeQS = models.FormKitSchemaNode.objects
    nodes = objects.from_change(latest_change)
    response["latest_change"] = (
        nodes.aggregate(_=Max("track_change"))["_"] or latest_change
    )
    add_never_cache_headers(response)
    return node_queryset_response(nodes)


@router.get(
    "list-related-nodes",
    response=list[NodeChildrenOut],
    exclude_defaults=True,
    exclude_none=True,
)
def get_related_nodes(request, response: HttpResponse, latest_change: int | None = -1):
    """
    Get all of the FormKit node relationships in the database
    """
    add_never_cache_headers(response)
    objects: models.NodeChildrenManager = models.NodeChildren.objects
    return objects.aggregate_changes_table(latest_change=latest_change)


@router.get(
    "list-components",
    response=list[FormComponentsOut],
    exclude_defaults=True,
    exclude_none=True,
    by_alias=True,
)
def get_components(request):
    values = models.FormComponents.objects.all()
    return values


@router.get(
    "schema/by-uuid/{schema_id}",
    response=formkit_schema.FormKitSchema,
    exclude_none=True,
    by_alias=True,
)
def get_schemas(request, schema_id: UUID):
    """
    Get a schema based on its UUID
    """
    schema: models.FormKitSchema = get_object_or_404(
        models.FormKitSchema.objects, id=schema_id
    )
    model = schema.to_pydantic()
    return model


@router.get(
    "schema/all",
    response=list[formkit_schema.FormKitSchema],
    exclude_none=True,
    by_alias=True,
)
def get_all_schemas(request):
    """
    Get all schemas
    """
    schemas = models.FormKitSchema.objects.all()
    model = [s.to_pydantic() for s in schemas]
    return model


@router.get(
    "schema/by-label/{label}",
    response=formkit_schema.FormKitSchema,
    exclude_none=True,
    by_alias=True,
)
def get_schema_by_label(request, label: str):
    """
    Get a schema based on its label
    """
    schema: models.FormKitSchema = get_object_or_404(
        models.FormKitSchema.objects, label=label
    )
    model = schema.to_pydantic()
    return model


@router.get(
    "node/{node_id}",
    response=formkit_schema.FormKitNode,
    exclude_none=True,
    by_alias=True,
)
def get_node(request, node_id: UUID):
    """
    Gets a node based on its UUID
    """
    node: models.FormKitSchemaNode = get_object_or_404(
        models.FormKitSchemaNode.objects, id=node_id
    )
    instance = node.get_node()
    return instance


@router.get("/options", response=list[Option], exclude_none=True)
def list_options(request: HttpRequest, response: HttpResponse):
    """
    List all available "native" FormKit ninja labels and links
    """
    return models.Option.objects.annotate(group_name=F("group__group"))


class FormKitErrors(BaseModel):
    errors: list[str] = []
    field_errors: dict[str, str] = {}


@router.delete("delete", response=NodeInactiveType, exclude_none=True, by_alias=True)
def delete_node(request, node_id: UUID):
    """
    Delete a node based on its UUID
    """
    with transaction.atomic():
        node: models.FormKitSchemaNode = get_object_or_404(
            models.FormKitSchemaNode.objects, id=node_id
        )
        node.delete()
        # node.refresh_from_db()
        objects: models.NodeQS = models.FormKitSchemaNode.objects
        return node_queryset_response(objects.filter(pk=node_id))[0]


class FormKitNodeIn(Schema):
    """
    Creates a new FormKit text or number node
    We'd like to use `formkit_schema.FormKitSchemaFormKit`
    here but that `discriminated node` stuff makes it hard
    """

    formkit: str = Field(default="text", alias="$formkit")
    label: str | None = None
    key: str | None = None
    name: str | None = None
    placeholder: str | None = None
    help: str | None = None

    # Fields from "number"
    max: int | None = None
    min: int | None = None
    step: str | None = None

    # Field from dropdown/select/autocomplete/radio/checkbox
    options: str | None = None

    # Used for Creates
    parent_id: UUID | None = None

    # Used for Updates
    uuid: UUID = Field(default_factory=uuid4)

    # Used for "Add Group"
    # This should include an `icon`, `title` and `id` for the second level group
    additional_props: dict[str, str | int] | None = None

    @cached_property
    def parent(self):
        if self.parent_id is None:
            return None, None
        try:
            parent = models.FormKitSchemaNode.objects.get(pk=self.parent_id)
        except models.FormKitSchemaNode.DoesNotExist:
            return None, ["The parent node given does not exist"]
        if parent.node.get("$formkit") not in {"group", "repeater"}:
            return None, ["The parent node given is not a group or repeater"]
        if parent is None:
            return None, ["The parent node given does not exist"]
        return parent, None

    @cached_property
    def parent_names(self) -> set[str]:
        """
        Return the names of parent nodes' child nodes.
        The saved child node must not use any of these names.
        """
        parent, parent_errors = self.parent
        if self.parent[0] and self.child:
            # Ensures that names are not "overwritten"
            return set(
                parent.children.exclude(pk=self.child.pk).values_list(
                    "node__name", flat=True
                )
            )
        elif self.parent[0]:
            return set(parent.children.values_list("node__name", flat=True))
        else:
            return set()

    @cached_property
    def child(self):
        # The uuid may belong to a node or may be a new value
        try:
            return models.FormKitSchemaNode.objects.get(pk=self.uuid)
        except models.FormKitSchemaNode.DoesNotExist:
            return models.FormKitSchemaNode(pk=self.uuid, node={})

    @cached_property
    def preferred_name(self):
        """
        Fetch a suitable name for the database to use.
        This name must be unique to the 'parent' group, a valid Python id, valid Django id,
        preferably lowercase.
        """
        # If "name" is not provided use the "label" field
        if self.name is not None:
            return disambiguate_name(make_name_valid_id(self.name), self.parent_names)
        elif self.label is not None:
            return disambiguate_name(make_name_valid_id(self.label), self.parent_names)
        return make_name_valid_id(f"{uuid4().hex[:8]}_unnamed")

    model_config = ConfigDict(populate_by_name=True, ignored_types=(cached_property,))


def create_or_update_child_node(payload: FormKitNodeIn):
    parent, parent_errors = payload.parent
    child = payload.child

    if parent_errors:
        return None, parent_errors
    if child.is_active is False:
        return None, ["This node has already been deleted and cannot be edited"]

    values = payload.dict(
        by_alias=True,
        exclude_none=True,
        exclude={"parent_id", "uuid"}
        | {"parent", "child", "preferred_name", "parent_names"},
    )
    # Ensure the name is unique and suitable
    # Do not replace existing names though
    existing_name = (
        child.node.get("name", None) if isinstance(child.node, dict) else None
    )
    if existing_name is None:
        values["name"] = payload.preferred_name
    child.node.update(values)
    if payload.additional_props is not None:
        child.node.update(payload.additional_props)

    child.label = payload.label

    if isinstance(payload.additional_props, dict):
        if label := payload.additional_props.get("label"):
            # groups require a label
            child.label = label

    child.node_type = "formkit"

    with transaction.atomic():
        child.save()
        if parent:
            models.NodeChildren.objects.create(parent=parent, child=child)

    return child, []


def make_name_valid_id(in_: str):
    """
    Take a string. Replace any python-invalid characters with '_'
    """
    subbed = re.sub(r"\W|^(?=\d)", "_", in_)
    while subbed[-1] == "_":
        subbed = subbed[:-1]
    return subbed.lower()


def disambiguate_name(name_in: str, used_names: Sequence[str]):
    suffix = 1
    if name_in not in used_names:
        return name_in
    while f"{name_in}_{suffix}" in used_names:
        suffix = suffix + 1
    return f"{name_in}_{suffix}"


@router.post(
    "create_or_update_node",
    response={
        HTTPStatus.OK: NodeReturnType,
        HTTPStatus.INTERNAL_SERVER_ERROR: FormKitErrors,
    },
    exclude_none=True,
    by_alias=True,
)
def create_or_update_node(request, response: HttpResponse, payload: FormKitNodeIn):
    """
    Creates or updates a node in the FormKitSchemaNode model.

    This function takes payload of type FormKitNodeIn.
    It fetches the parent node if it exists and checks if it is a group or repeater.
    If the parent node is not valid or is not a group or repeater, it returns an error response.
    Otherwise, it proceeds to create or update the node.

    Args:
        request: The request object.
        response (HttpResponse): The HttpResponse object.
        payload (FormKitNodeIn): The payload containing the data for the node to be created or updated.

    Returns:
        HTTPStatus: The status of the HTTP response.
        FormKitErrors: The errors encountered during the process, if any.
    """

    error_response = FormKitErrors()
    # Update the payload "name"
    # When label is provided, use the label to generate the name
    # Fetch parent node, if it exists, and check that it is a group or repeater

    try:
        child, errors = create_or_update_child_node(payload)
        if errors:
            error_response.errors.append(errors)
    except Exception as E:
        error_response.errors.append(f"{E}")

    if error_response.errors or error_response.field_errors:
        return HTTPStatus.INTERNAL_SERVER_ERROR, error_response

    return node_queryset_response(
        models.FormKitSchemaNode.objects.filter(pk__in=[child.pk])
    )[0]



from formkit_ninja.html_parser import FormKitTagParser


def test_basic_example():
    html_content = """
    <formkit type="text" />
    """
    parser = FormKitTagParser(html_content)
    assert parser.tags[0].model_dump(exclude_none=True, by_alias=True) == {
        "$formkit": "text"
    }



from __future__ import annotations

import logging
import operator
import warnings
from collections import Counter
from functools import reduce

import django.core.exceptions
from django import forms
from django.contrib import admin
from django.db.models import JSONField
from django.http import HttpRequest

from formkit_ninja import formkit_schema, models

logger = logging.getLogger(__name__)


# Define fields in JSON with a tuple of fields
# The key of the dict provided is a JSON field on the model
JsonFieldDefn = dict[str, tuple[str | tuple[str, str], ...]]


class ItemAdmin(admin.ModelAdmin):
    list_display = ("name",)


class JsonDecoratedFormBase(forms.ModelForm):
    """
    Adds additional fields to the admin where a model has a JSON field
    and some appropriate (tbc?) field parameters
    """

    # extra = forms.CharField(label="Extra", max_length=128, required=False)
    # hello_world = forms.CharField(widget=forms.NumberInput, required=False)

    # key is the name of a `models.JSONField` on the model
    # value is a list of fields to get/set in that JSON field
    _json_fields: JsonFieldDefn = {
        "my_json_field": ("formkit", "description", "name", "key", "id")
    }

    def get_json_fields(self) -> JsonFieldDefn:
        """
        Custom which json fields will be included in the return
        """
        return self._json_fields

    def _field_check(self):
        """
        Check that JSON fields specified are json fields on the model
        do not clash with model fields
        """

        def check_json_fields_exist():
            """
            Check that JSON fields specified are json fields on the model
            """
            for field in self.get_json_fields().keys():
                try:
                    model_field = self.Meta.model._meta.get_field(field)
                    if not isinstance(model_field, JSONField):
                        raise KeyError(
                            f"Expected a JSONField named {field} on the model"
                        )
                except django.core.exceptions.FieldDoesNotExist as E:
                    raise KeyError(
                        f"Expected a JSONField named {field} on the model"
                    ) from E

        def check_no_duplicates():
            """
            Checks that the `_json_fields` specified do not clash with fields
            on the model
            """
            fields_in_model = Counter(self.Meta.fields)

            # These are all the fields we've "JSON"ified
            for json_keys in self.get_json_fields().values():
                fields_in_model.update((k for k in json_keys if isinstance(k, str)))
                fields_in_model.update(
                    (k[0] for k in json_keys if not isinstance(k, str))
                )

            # Duplicate fields raise an exception
            duplicates = list(((k, v) for k, v in fields_in_model.items() if v > 1))
            if duplicates:
                raise KeyError(f"Some fields were duplicated: {','.join(duplicates)}")

        check_json_fields_exist()
        check_no_duplicates()

    def _set_json_fields(self, instance):
        """
        Assign JSON field content to Form fields
        """
        for field, keys in self.get_json_fields().items():
            # Extract the dict of JSON values from the model instance if supplied
            values = getattr(instance, field, {}) or {}  # Don't allow none:
            # field_name is the property on the ModelForm to use.
            # This allows "aliasing" so that fields on the model / JSON are not shadowed.

            fields_from_json = set()

            for key in keys:
                if isinstance(key, str):
                    form_field = key
                    json_field = key
                else:
                    form_field, json_field = key
                fields_from_json.add(json_field)
                field = self.fields.get(form_field)
                if not field:
                    warnings.warn(f"The field {form_field} was not found on the form")
                else:
                    # The value, extracted from the JSON value in the database
                    # If there's an underscore do a Django-style "lookup"
                    # e.g. `formkit__name` -> `formkit["name"]`
                    if "__" in json_field:
                        nested_field_name = json_field.split("__")[0]
                        if nested_field_name in values:
                            field_value = values[nested_field_name].get(
                                json_field.split("__")[1], None
                            )
                    field_value = values.get(json_field, None)
                    # The initial value of the admin form is set to the value of the JSON attrib
                    field = self.fields.get(form_field).initial = field_value

            # Here we can warn if there are any "hidden" JSON fields
            # Todo: Handle 'nested' values such as `attrs__classes`
            if missing := list(set(values) - fields_from_json - {"node_type"}):
                warnings.warn(f"Some JSON fields were hidden: {','.join(missing)}")
                warnings.warn(
                    f"Consider adding fields {missing} to {self.__class__.__name__}"
                )

    def __init__(self, *args, **kwargs):
        """
        When the form is initialized, additional "Fields"
        can be populated from a JSON data field.
        """
        super().__init__(*args, **kwargs)
        if instance := kwargs["instance"]:
            self._set_json_fields(instance)

    def save(self, commit=True):
        """
        Updates the JSON field(s) from the fields specified in the `_json_fields` dict
        """

        for field, keys in self.get_json_fields().items():
            # Populate a JSON field in a model named "form"
            # from a set of standard form elements
            data = {}
            for key in keys:
                if isinstance(key, str):
                    form_field = key
                    json_field = key
                else:
                    form_field, json_field = key
                if field_value := self.cleaned_data.get(form_field, None):
                    if "__" in json_field:
                        if json_field.split("__")[0] in data:
                            data[json_field.split("__")[0]][
                                json_field.split("__")[1]
                            ] = field_value
                        else:
                            data[json_field.split("__")[0]] = {
                                json_field.split("__")[1]: field_value
                            }
                    else:
                        data[json_field] = field_value
        setattr(self.instance, field, data)
        return super().save(commit=commit)


class NewFormKitForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "node_type", "description")


class OptionForm(forms.ModelForm):
    class Meta:
        model = models.Option
        exclude = ()


class FormComponentsForm(forms.ModelForm):
    class Meta:
        model = models.FormComponents
        exclude = ()


class FormKitSchemaComponentInline(admin.TabularInline):
    model = models.FormComponents
    readonly_fields = (
        "node",
        "created_by",
        "updated_by",
    )
    ordering = ("order",)
    extra = 0


class FormKitNodeGroupForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "description", "additional_props", "is_active", "protected")

    _json_fields = {
        "node": ("name", ("formkit", "$formkit"), "if_condition", ("html_id", "id")),
    }
    html_id = forms.CharField(
        required=False,
        help_text="Use this ID if adding conditions to other fields (hint: $get(my_field).value === 8)",
    )
    name = forms.CharField(required=False)
    formkit = forms.ChoiceField(
        required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES, disabled=True
    )
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitNodeForm(JsonDecoratedFormBase):
    """
    This is the most common component type: the 'FormKit' schema node
    """

    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "description",
            "additional_props",
            "option_group",
            "is_active",
            "protected",
        )

    # The `_json_fields["node"]` item affects the admin form,
    # adding the fields included in the `FormKitSchemaProps.__fields__.items` dict
    _json_fields = {
        "node": (
            ("formkit", "$formkit"),
            "name",
            "key",
            "if_condition",
            "options",
            ("node_label", "label"),
            "placeholder",
            "help",
            # Validation of fields
            "validation",
            "validationLabel",
            "validationVisibility",
            "validationMessages",
            "prefixIcon",
            # Numeric forms
            "min",
            "max",
            "step",
            ("html_id", "id"),
            ("onchange", "onChange"),
        )
    }

    formkit = forms.ChoiceField(
        required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES
    )
    name = forms.CharField(required=False)
    if_condition = forms.CharField(widget=forms.TextInput, required=False)
    key = forms.CharField(required=False)
    label = forms.CharField(required=False)
    node_label = forms.CharField(required=False)
    placeholder = forms.CharField(required=False)
    help = forms.CharField(required=False)
    html_id = forms.CharField(
        required=False,
        help_text="Use this ID if adding conditions to other fields (hint: $get(my_field).value === 8)",
    )
    onchange = forms.CharField(
        required=False,
        help_text="Use this to trigger a function when the value of the field changes",
    )
    options = forms.CharField(
        required=False,
        help_text="Use this if adding Options using a JS function (hint: $get(my_field).value )",
    )
    validation = forms.CharField(required=False)
    validationLabel = forms.CharField(required=False)
    validationVisibility = forms.CharField(required=False)
    validationMessages = forms.JSONField(required=False)
    validationRules = forms.CharField(
        required=False,
        help_text="A function for validation passed into the schema: a key on `formSchemaData`",
    )
    prefixIcon = forms.CharField(required=False)

    # NumberNode props

    max = forms.IntegerField(required=False)
    min = forms.IntegerField(required=False)
    step = forms.IntegerField(required=False)


class FormKitNodeRepeaterForm(FormKitNodeForm):
    def get_json_fields(self) -> JsonFieldDefn:
        return {
            "node": (
                *(super()._json_fields["node"]),
                "addLabel",
                "upControl",
                "downControl",
                "itemsClass",
                "itemClass",
            )
        }

    addLabel = forms.CharField(required=False)
    upControl = forms.BooleanField(required=False)
    downControl = forms.BooleanField(required=False)
    itemsClass = forms.CharField(required=False)
    itemClass = forms.CharField(required=False)
    max = forms.IntegerField(required=False)
    min = forms.IntegerField(required=False)


class FormKitTextNode(forms.ModelForm):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "description", "text_content", "is_active", "protected")


class FormKitElementForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "description", "text_content", "is_active", "protected")

    _skip_translations = {"label", "placeholder"}
    _json_fields = {"node": (("el", "$el"), "name", "if_condition", "attrs__class")}

    el = forms.ChoiceField(
        required=False, choices=models.FormKitSchemaNode.ELEMENT_TYPE_CHOICES
    )
    name = forms.CharField(
        required=False,
    )
    attrs__class = forms.CharField(
        required=False,
    )
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitConditionForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        # fields = '__all__'
        fields = (
            "label",
            "description",
        )

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}

    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )
    then_condition = forms.CharField(
        max_length=256,
        required=False,
    )
    else_condition = forms.CharField(
        max_length=256,
        required=False,
    )


class FormKitComponentForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "description", "is_active", "protected")

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}


class NodeChildrenInline(admin.TabularInline):
    """
    Nested HTML elements
    """

    model = models.NodeChildren
    fields = ("child", "order", "track_change")
    ordering = ("order",)
    readonly_fields = ("track_change", "child")
    fk_name = "parent"
    extra = 0


class NodeParentsInline(admin.TabularInline):
    """
    Nested HTML elements
    """

    model = models.NodeChildren
    fields = ("parent", "order", "track_change")
    ordering = ("order",)
    readonly_fields = ("track_change", "parent")
    fk_name = "child"
    extra = 0


class NodeInline(admin.StackedInline):
    """
    Nodes related to Option Groups
    """

    model = models.FormKitSchemaNode
    fields = ("label", "node_type", "description")
    extra = 0


class SchemaLabelInline(admin.TabularInline):
    model = models.SchemaLabel
    extra = 0


class SchemaDescriptionInline(admin.TabularInline):
    model = models.SchemaDescription
    extra = 0


class FormKitSchemaForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchema
        exclude = ("name",)


@admin.register(models.FormKitSchemaNode)
class FormKitSchemaNodeAdmin(admin.ModelAdmin):
    list_display = (
        "label",
        "is_active",
        "id",
        "node_type",
        "option_group",
        "formkit_or_el_type",
        "track_change",
        "key_is_valid",
        "protected",
    )
    list_filter = ("node_type", "is_active", "protected")
    readonly_fields = ("track_change",)
    search_fields = ["label", "description", "node", "node__el"]

    @admin.display(boolean=True)
    def key_is_valid(self, obj) -> bool:
        """
        If it's a Formkit type, check that its
        key is suitable for python + django
        """
        if not obj.node:
            return True
        if not isinstance(obj.node, dict):
            return True
        if "name" not in obj.node:
            return True
        try:
            key = obj.node.get("name")
            if not isinstance(key, str):
                raise TypeError
            models.check_valid_django_id(key)
        except TypeError:
            return False
        return True

    def formkit_or_el_type(self, obj):
        if obj and obj.node and obj.node_type == "$formkit":
            return obj.node.get("$formkit", None)
        if obj and obj.node and obj.node_type == "$el":
            return obj.node.get("$el", None)

    def get_inlines(self, request, obj: models.FormKitSchemaNode | None):
        if not obj:
            return []
        return [NodeChildrenInline, NodeParentsInline]

    # # Note that although overridden these are necessary
    inlines = [NodeChildrenInline, NodeParentsInline]

    def get_fieldsets(
        self, request: HttpRequest, obj: models.FormKitSchemaNode | None = None
    ):
        """
        Set fieldsets to control the layout of admin “add” and “change” pages.
        fieldsets is a list of two-tuples, in which each two-tuple represents a <fieldset>
        on the admin form page. (A <fieldset> is a “section” of the form.)
        """
        fieldsets: list[tuple[str, dict]] = []
        if not getattr(obj, "node_type", None):
            warnings.warn("Expected a 'Node' with a 'NodeType' in the admin form")

        if not obj:
            # This default form is returned before a field
            # type is selected
            return super().get_fieldsets(request, obj)
        try:
            node = obj.get_node()
        except Exception as E:
            warnings.warn(f"{E}")
            return fieldsets

        if isinstance(node, formkit_schema.FormKitSchemaProps) and not isinstance(
            node,
            (
                formkit_schema.GroupNode,
                formkit_schema.FormKitSchemaDOMNode,
                formkit_schema.RepeaterNode,
            ),
        ):
            # Field validation applies to formkit
            # except repeater and group
            fieldsets.append(
                (
                    "Field Validation",
                    {
                        "fields": (
                            "validation",
                            "validationLabel",
                            "validationVisibility",
                            "validationMessages",
                            "validationRules",
                        )
                    },
                )
            )
        elif isinstance(node, formkit_schema.RepeaterNode):
            if obj.node and obj.node.get("$formkit", None) == "repeater":
                fieldsets.append(
                    (
                        "Repeater field properties",
                        {
                            "fields": (
                                "addLabel",
                                "upControl",
                                "downControl",
                                "itemsClass",
                                "itemClass",
                            )
                        },
                    )
                )

        grouped_fields = reduce(
            operator.or_, (set(opts["fields"]) for _, opts in fieldsets), set()
        )
        # Add 'ungrouped' fields
        fieldsets.insert(
            0,
            (
                None,
                {
                    "fields": [
                        field
                        for field in self.get_fields(request, obj)
                        if field not in grouped_fields
                    ]
                },
            ),
        )
        logger.info(fieldsets)
        return fieldsets

    def get_form(
        self,
        request: HttpRequest,
        obj: models.FormKitSchemaNode | None,
        change: bool | None = None,
        **kwargs,
    ):
        if not obj:
            return NewFormKitForm
        try:
            get_node = obj.get_node()
            for node_type, form_type in (
                (str, FormKitTextNode),
                (formkit_schema.GroupNode, FormKitNodeGroupForm),
                (formkit_schema.RepeaterNode, FormKitNodeRepeaterForm),
                (formkit_schema.FormKitSchemaDOMNode, FormKitElementForm),
                (formkit_schema.FormKitSchemaComponent, FormKitComponentForm),
                (formkit_schema.FormKitSchemaCondition, FormKitConditionForm),
                (formkit_schema.FormKitSchemaProps, FormKitNodeForm),
            ):
                if isinstance(get_node, node_type):
                    return form_type

        except Exception as E:
            warnings.warn(f"{E}")
            raise

        warnings.warn(f"Unable to determine form type for {obj}")
        return super().get_form(request, obj, change, **kwargs)


@admin.register(models.FormKitSchema)
class FormKitSchemaAdmin(admin.ModelAdmin):
    form = FormKitSchemaForm

    def get_inlines(self, request, obj: models.FormKitSchema | None):
        """
        For a "new object" do not show the Form Components
        """
        return (
            [
                SchemaLabelInline,
                SchemaDescriptionInline,
                FormKitSchemaComponentInline,
            ]
            if obj
            else [
                SchemaLabelInline,
                SchemaDescriptionInline,
            ]
        )

    inlines = [
        SchemaLabelInline,
        SchemaDescriptionInline,
        FormKitSchemaComponentInline,
    ]


@admin.register(models.FormComponents)
class FormComponentsAdmin(admin.ModelAdmin):
    list_display = (
        "label",
        "schema",
        "node",
        "order",
    )


class OptionLabelInline(admin.TabularInline):
    model = models.OptionLabel
    extra = 0


class OptionInline(admin.TabularInline):
    model = models.Option
    extra = 0
    fields = ("group", "object_id", "value", "order")
    readonly_fields = ("group", "object_id", "value")


@admin.register(models.Option)
class OptionAdmin(admin.ModelAdmin):
    list_display = (
        "object_id",
        "value",
        "order",
        "group",
    )
    inlines = [OptionLabelInline]
    list_select_related = ("group",)
    readonly_fields = ("group", "object_id", "value", "created_by", "updated_by")


@admin.register(models.OptionGroup)
class OptionGroupAdmin(admin.ModelAdmin):
    inlines = [OptionInline, NodeInline]


@admin.register(models.OptionLabel)
class OptionLabelAdmin(admin.ModelAdmin):
    list_display = (
        "label",
        "lang",
    )
    readonly_fields = ("option",)
    search_fields = ("label",)



# Generated by Django 4.2.5 on 2023-10-01 11:43

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models

from formkit_ninja import triggers


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0018_remove_formcomponents_order_on_insert_option_and_more"),
    ]

    operations = [
        # Note that this was added to create the sequence
        migrations.RunSQL(*triggers.create_sequence_migration(triggers.NODE_CHANGE_ID)),
        migrations.AddField(
            model_name="formkitschemanode",
            name="track_change",
            field=models.BigIntegerField(blank=True, null=True),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="version_on_update",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func="NEW.\"track_change\" = nextval('formkitschemanode_change_id'); RETURN NEW;",
                    hash="23d8a7e6e73b7b06f75c07e679c192634ec2a6d9",
                    operation="UPDATE OR INSERT",
                    pgid="pgtrigger_version_on_update_09b77",
                    table="formkit_ninja_formkitschemanode",
                    when="BEFORE",
                ),
            ),
        ),
    ]



from typing import Any

from pydantic import (BaseModel, RootModel, ValidatorFunctionWrapHandler,
                      field_validator)


class MyModel(BaseModel):
    value: str

    @field_validator("value", mode="wrap")
    @classmethod
    def translate(cls, value: Any, handler: ValidatorFunctionWrapHandler) -> str:
        return handler(value)[::-1]


def test_mymodel():
    model = MyModel.model_validate({"value": "This is a message"})
    assert model.value == "egassem a si sihT"


class SomeRootModel(RootModel):
    root: MyModel


def test_rootmodel():
    model = SomeRootModel.model_validate({"value": "This is a message"})
    assert model.root.value == "egassem a si sihT"



# Generated by Django 4.1.5 on 2023-04-26 08:00

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="FormComponents",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(
                        db_index=True, editable=False, verbose_name="order"
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "label",
                    models.CharField(
                        default=uuid.uuid4,
                        help_text="Used as a human-readable label",
                        max_length=1024,
                        unique=True,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("schema", "order"),
            },
        ),
        migrations.CreateModel(
            name="FormKitSchemaNode",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "node_type",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("$cmp", "Component"),
                            ("text", "Text"),
                            ("condition", "Condition"),
                            ("$formkit", "FormKit"),
                            ("$el", "Element"),
                            ("raw", "Raw JSON"),
                        ],
                        max_length=256,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        blank=True,
                        help_text="Decribe the type of data / reason for this component",
                        max_length=4000,
                        null=True,
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        help_text="Used as a human-readable label",
                        max_length=1024,
                        unique=True,
                    ),
                ),
                (
                    "node",
                    models.JSONField(
                        blank=True,
                        help_text="A JSON representation of select parts of the FormKit schema",
                        null=True,
                    ),
                ),
                (
                    "additional_props",
                    models.JSONField(
                        blank=True,
                        help_text="User space for additional, less used props",
                        null=True,
                    ),
                ),
                (
                    "translation_context",
                    models.CharField(
                        blank=True,
                        help_text="The gettext context to use to translate Options, label, placeholder and help text",
                        max_length=1024,
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Translatable",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "object_id",
                    models.UUIDField(
                        editable=False,
                        help_text="The UUID of the model which this translation relates to",
                    ),
                ),
                ("language_code", models.CharField(editable=False, max_length=3)),
                (
                    "field",
                    models.CharField(
                        editable=False,
                        help_text="The field on the generic model to translate",
                        max_length=100,
                    ),
                ),
                (
                    "value",
                    models.CharField(
                        blank=True, editable=False, max_length=5000, null=True
                    ),
                ),
                (
                    "context",
                    models.CharField(
                        blank=True, editable=False, max_length=1024, null=True
                    ),
                ),
                ("msgstr", models.CharField(help_text="The string", max_length=5000)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Option",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(
                        db_index=True, editable=False, verbose_name="order"
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                ("value", models.CharField(max_length=1024)),
                ("label", models.CharField(max_length=1024)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "field",
                    models.ForeignKey(
                        limit_choices_to={"node__formkit__in": ["select", "radio"]},
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="NodeChildren",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(
                        db_index=True, editable=False, verbose_name="order"
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "child",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="parent",
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        limit_choices_to={"node__node_type": "$el"},
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
            },
        ),
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "order",
                    models.PositiveIntegerField(
                        db_index=True, editable=False, verbose_name="order"
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        limit_choices_to={"node__formkit": "group"},
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "member",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="members",
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="children",
            field=models.ManyToManyField(
                blank=True,
                through="formkit_ninja.NodeChildren",
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="group",
            field=models.ManyToManyField(
                blank=True,
                through="formkit_ninja.Membership",
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="FormKitSchema",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated", models.DateTimeField(auto_now=True, null=True)),
                ("key", models.SlugField(max_length=1024, unique=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "nodes",
                    models.ManyToManyField(
                        through="formkit_ninja.FormComponents",
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="formcomponents",
            name="node",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.AddField(
            model_name="formcomponents",
            name="schema",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.formkitschema",
            ),
        ),
        migrations.AddField(
            model_name="formcomponents",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
